{--------------------------------------------------------------------}
{ TmwDList V 1.0                                                     }
{ Written by Martin Waldenburg 1996.                                 }
{ Copyright by Martin Waldenburg 1996. All rights reserved.          }
{ This is FreeWare.                                                  }
{ Not testet with Delphi1.                                           }
{ It's provided as is, without a warranty of any kind.               }
{ You use it at your own risc.                                       }
{ You may use and distribute it freely.                              }
{ But You may not say it's your work                                 }
{ If you distribute it you must provide all Files.                   }
{--------------------------------------------------------------------}
{ This Component demonstrates the implementation and use of a        }
{ sortable doubly linked list using MergeSort.                                }
{ I couldn't find a doubly linked list implemented as a component,   }
{ I wrote it myself.                                                 }
{ I implemented this with MergeSort because I've wasted so mucht     }
{ time waiting for sorting.                                          }
{ Mergesort is most efficient if used with linked lists, in which    }
{ case it comes very close to the theoretically most efficient       }
{ sorting algorithm possible.                                        }
{ This is only a simple implementation, you can double the speed     }
{ verry easy.  However it's fast, fast, fast.                        }
{ Take care only allocating enough memory that Windows doesn't swap. }
{ In this case you must sort parts, save to disc, then merge the     }
{ disc files                                                         }
{ Enjoy it's free.                                                   }
{--------------------------------------------------------------------}

unit mwDList;

interface

uses
  SysUtils, Classes;

type
  PNode = ^TNode;
  TNode = record
            next,
            prev : PNode;
            data   : string;
          end;

type
  TmwDList = class(TComponent)
  private
    Node,
    head,      { Head and Bottom are pseude Nodes, they don't contain any Data. }
    bottom,    { They make it much easier to manage the list. }
    current,
    todoA,
    todoB: PNode;
    a_count,
    b_count,
    FSize : Integer;
    FSorted,
    FIgnore,
    FReversed : boolean;
    Function CompareFunction(a, b: PNode):Integer;
    Function CreateNode(n, p:PNode; item:String):PNode;
    Procedure PlaceBefore(a, b, c: PNode);
    Procedure Merge(a, b: PNode);
    Procedure MergeSort;
  protected
  public
    constructor Create(AOwner:TComponent);override;
    destructor Destroy;override;
    Property Size : integer read FSize write FSize ;
    Procedure Delete(aNode: PNode);
    Procedure DeleteEmptyNodes;
    Procedure Find(item: String);
    Function GetData : string;
    Procedure InsertFirst(item: String);
    Procedure InsertBefore(item: String);
    Procedure InsertBehind(item: String);
    Procedure InsertLast(item: String);
    Procedure MoveFirst;
    Procedure MoveLast;
    Procedure MoveNext;
    Procedure MovePrevious;
    Procedure LoadFromFile(FileName: String);
    Procedure SaveToFile(FileName: String);
    Procedure Sort;
  published
    Property Sorted : boolean read FSorted write FSorted ;
    Property IgnoreCase : boolean read FIgnore write FIgnore ;
    Property Reversed : boolean read FReversed write FReversed ;
    Property Name;
    Property Tag;
end;

procedure Register;


implementation


procedure Register;
begin
  RegisterComponents('mw',[TmwDList]);
end;

{ Create a node with given values}
Function TmwDList.CreateNode(n, p:PNode; item:String):PNode;
var
  newNode: PNode;
begin
  new(Node);
  newNode:= Node;
  with newNode^ do begin
       next:= n;
       prev:= p;
       data:= item;
  end;
    Result:= newNode;
end;


{ Create the list with default values }
constructor TmwDList.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);
  head:= CreateNode(nil, nil, 'this is the head node');
  bottom:= CreateNode(nil, head, 'this is the bottom node');
  head^.next:= bottom;
  head^.prev:= head;
  bottom^.next:= bottom;
  current:= head;
  todoA:= head;
  todoB:= head;
  a_count:= 0;
  b_count:= 0;
  FSize:= 0;
  FSorted:= False;
  FIgnore:= False;
  FReversed:= False;
end;
 

{ Destroy the whole list, node by node }
destructor TmwDList.Destroy;
begin
   Node := head;
   while Node <> bottom do begin
         head:= Node;
         Node := head^.Next;
         Dispose(head);
   end;
   Dispose(Node);
  inherited Destroy;
end;
{ Compare the data of two nodes }
Function TmwDList.CompareFunction(a, b: PNode):Integer;
begin
  If FIgnore = False then
     Result:= CompareStr(a^.data, b^.data)
  else
     Result:= CompareText(a^.data, b^.data);
end;

{ Delete a given node }
Procedure TmwDList.Delete(aNode: PNode);
begin
   if (not(Fsize = 0)) and (not(aNode = head)) and (not(aNode = bottom)) then
     begin
       current:= aNode^.prev;
       current^.next:= aNode^.next;
       aNode^.next^.prev:= current;
       Dispose(aNode);
       dec(Fsize);
     end;
end;

{ Delete empty nodes }
Procedure TmwDList.DeleteEmptyNodes;
begin
  MoveFirst;
  while (not(current = bottom)) do
    begin
      if current^.data = '' then
        begin
          Delete(current);
        end else
        MoveNext;
    end;
end;

{ Look for a item in the list. Starts at head node. }
Procedure TmwDList.Find(item: String);
begin
  MoveFirst;
  while (not(current^.data = item)) and (not(current = bottom)) do
    begin
      MoveNext;
    end;
end;

 
{ Retrive data of current node. }
function TmwDList.GetData: String;
begin
    Result := current^.data;
end;

{ Insert as first node, make new node current }
Procedure TmwDList.InsertFirst(item: String);
begin
  Node:= CreateNode(current, head, item);
  head^.next:= Node;
  current^.prev:= Node;
  current:= Node;
  inc(Fsize);
end;

{ Insert before current node, make new node current }
Procedure TmwDList.InsertBefore(item: String);
begin
  Node:= CreateNode(current, current^.prev, item);
  current^.prev^.next:= Node;
  current^.prev:= Node;
  current:= Node;
  inc(Fsize);
end;

{ Insert behind current node, make new node current }
Procedure TmwDList.InsertBehind(item: String);
begin
  Node:= CreateNode(current^.next, current, item);
  current^.next^.prev:= Node;
  current^.next:= Node;
  current:= Node;
  inc(Fsize);
end;

{ insert Node after current item, make newNode current }
procedure TmwDList.InsertLast(item: String);
begin
  Node:= CreateNode(bottom, current, item);
  current^.next:= Node;
  bottom^.prev:= Node;
  current:= Node;
  inc(Fsize);
end;


{ make first value in list current }
procedure TmwDList.MoveFirst;
begin
    current := head^.next;
end;

{ make last value in list current }
procedure TmwDList.MoveLast;
begin
    current := bottom^.prev;
end;

{ make next value in list current }
procedure TmwDList.MoveNext;
begin
    current := current^.next;
end;

{ make previous value in list current }
procedure TmwDList.MovePrevious;
begin
    current := current^.prev;
end;

{ Load the list from file }
Procedure TmwDList.LoadFromFile(FileName: String);
var
  FromFile: TextFile;
  TextBuffer: String;
begin
  AssignFile(FromFile, FileName);
  Reset(FromFile);
  current:= head;
  while not Eof(FromFile) do begin
    Readln(FromFile, TextBuffer);
    InsertLast(TextBuffer);
  end;
   CloseFile(FromFile);
end;
 

{ Save the list to file }
Procedure TmwDList.SaveToFile(FileName: String);
var
  ToFile: TextFile;
  TextBuffer: String;
begin
  AssignFile(ToFile, FileName);
  Rewrite(ToFile);
  if FReversed = False then
    begin
      MoveFirst;
      while not (current = bottom) do begin
        TextBuffer:= GetData;
        Writeln(ToFile, TextBuffer);
        MoveNext;
      end;
    end else
    begin
      MoveLast;
      while not (current = head) do begin
        TextBuffer:= GetData;
        Writeln(ToFile, TextBuffer);
        MovePrevious;
      end;
    end;
   CloseFile(ToFile);
end;
 

{ Place c before a }
Procedure TmwDList.PlaceBefore(a, b, c: PNode);
begin
  a^.prev^.next:= c;
  b^.prev:= c^.prev;
  c^.prev^.next:= b;
  c^.prev:= a^.prev;
  a^.prev:= c;
  c^.next:= a;
end;
 
 
{ Merge the sublists a and b }
Procedure TmwDList.Merge(a, b: PNode);
var
  c: PNode;
begin
  If CompareFunction(todoB^.prev, todoB) <= 0 then exit;
  while (not(a = todoB)) and (not(b = todoA)) do
    if CompareFunction(a, b) <= 0 then
      begin
        a:= a^.next;
      end
    else
      begin
        c:= b;
        todoB:= b^.next;
        b:= todoB;
        PlaceBefore(a, b, c);
      end;
end;
 

{ Main sorting routine }
Procedure TmwDList.MergeSort;
var a, b, run: PNode;
  i, N: integer;
begin
  N:=1;
  repeat
    todoA:= head^.next;
    repeat
      a_count:= 0;
      b_count:= 0;
      run:=todoA;
      a:= run; for i:=1 to N do begin run:=run^.next; a_count:= a_count +1;
      todoB:=run; end;


      b:= todoB;
      repeat
        begin run:=run^.next; b_count:= b_count +1;
        todoA:= run; end;
      until (b_count = N) or (todoA = bottom);
        Merge(a,b);
    until (Fsize = a_count + b_count) or (todoA^.next = bottom);
    N:=N+N;
  until Fsize = a_count + b_count;
end;
 
 
{~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~}

{ Start sorting }
Procedure TmwDList.Sort;
begin
  MergeSort;
end;

end.

{------------------------------------------------------------}
{ Martin Waldenburg                                          }
{ Landaeckerstrasse 27                                       }
{ 71642 Ludwigsburg                                          }
{ Germany                                                    }
{                                                            }
{ Sugestions, improvements and code of any kind are welcome  }
{ Thanks for trying my code.                                 }
{ Share your code!                                           }
{------------------------------------------------------------}
