{--------------------------------------------------------------------}
{ TSorArray V 1.2                                                    }
{ Written by Martin Waldenburg 1997.                                 }
{ Placed to the Public Domain.                                       }
{ Not testet with Delphi1.                                           }
{ It's provided as is, without a warranty of any kind.               }
{ You use it at your own risc.                                       }
{ You may use and distribute it freely.                              }
{--------------------------------------------------------------------}
{ This demonstrates the implementation and use of a sortable dynamic }
{ array using the three fastest sorting Algorithm available.         }
{ Enjoy it's free.                                                   }
{--------------------------------------------------------------------}

unit mwSArray;

{$R-}

interface

uses SysUtils, Windows;

type
  PDynArray = ^TDynArray;
  TDynArray = array[0..0] of Pointer;
  TSortArrayCompare = function (Item1, Item2: Pointer): Integer;


{ TSubArray definiert die Grenzen eines SubArrays und bestimmt ob das
  SubArray gefllt ist oder nicht.
  Der MergeSort Algorithmus wird durch diese Klasse leichter lesbar}
type
  TSubArray = class(TObject)
  private
    FMax: LongInt;
  protected
  public
    FLeft: LongInt;  { - Initialized to 0. }
    FRight: LongInt;  { - Initialized to 0. }
    Full: Boolean;
    constructor Create(MaxValue: LongInt);
    destructor Destroy; override;
    procedure Init(LeftEnd, RightEnd: LongInt);
    procedure Next;
  end;  { TSubArray }


type

{ TSorArray class }
  TSorArray = class(TObject)
  private
    FLeftArray, FRightArray: TSubArray;
    FSorArray, TempArray, SwapArray: PDynArray;
    FCount: Integer;
    fCapacity:Integer;
    procedure SetCapacity(NewCapacity:Integer);
    procedure Expand;
  protected
    function Get(Index: Integer): Pointer;
    procedure Put(Index: Integer; Item: Pointer);
    function Comb(jumpsize0: Integer; SorCompare: TSortArrayCompare): boolean;
    procedure Merge(SorCompare: TSortArrayCompare);
  public
    destructor Destroy; override;
    function Add(Item: Pointer): Integer;
    procedure Clear;
    function Last: Pointer;
    procedure CombSortW(SorCompare: TSortArrayCompare);
    procedure QuickSort(SorCompare: TSortArrayCompare);
    procedure MergeSort(SorCompare: TSortArrayCompare);
    property Count: Integer read FCount write FCount;
    property Items[Index: Integer]: Pointer read Get write Put; default;
    property SorArray: PDynArray read FSorArray;
    property Capacity:Integer read fCapacity write SetCapacity;
  published
  end;

implementation

procedure TSubArray.Next;
  begin
    inc(FLeft);
    if (FLeft > FRight) or (FLeft > FMax) then Full:= False;
  end;  { Next }

procedure TSubArray.Init(LeftEnd, RightEnd: LongInt);  { public }
  begin
    FLeft:= LeftEnd;
    FRight:= RightEnd;
    if FLeft > FMax then Full:= False else Full:= True;
  end;  { Init }

destructor TSubArray.Destroy;
  begin
    inherited Destroy;
  end;  { Destroy }

constructor TSubArray.Create(MaxValue: LongInt);
  begin
    FLeft := 0;
    FRight := 0;
    Full := False;
    FMax := MaxValue;
  end;  { Create }

{ TSorArray }
destructor TSorArray.Destroy;
begin
  Clear;
  inherited Destroy;
end;

function TSorArray.Add(Item: Pointer): Integer;
begin
  Result := FCount;
  if Result = FCapacity then Expand;
  FSorArray[Result] := Item;
  Inc(FCount);
end;

procedure TSorArray.Expand;
begin
  SetCapacity(FCapacity + 1024);
end;

procedure TSorArray.SetCapacity(NewCapacity:Integer);
begin
  FCapacity:= NewCapacity;
  ReallocMem(FSorArray, FCapacity * SizeOf(Pointer));
end;

procedure TSorArray.Clear;
begin
  FCount:= 0;
  ReallocMem(TempArray, 0);
  ReallocMem(FSorArray, FCount);
end;

function TSorArray.Get(Index: Integer): Pointer;
begin
  Result := FSorArray[Index];
end;

function TSorArray.Last: Pointer;
begin
  Result := Get(FCount - 1);
end;

procedure TSorArray.Put(Index: Integer; Item: Pointer);
begin
  FSorArray[Index] := Item;
end;

{ Multipication by 0.76 gives a slightly better result than
  division by 1.3.
  Because of the FOR loop it runs faster on arrays starting with one}
function TSorArray.Comb(jumpsize0: Integer; SorCompare: TSortArrayCompare): boolean;
const irgap: Real = 0.76;
var
  i, SwapNumber: Integer;
  Temp: Pointer;
begin
 repeat
    jumpsize0 := trunc(jumpsize0*irgap);
    SwapNumber:= jumpsize0+1;
    Result:= True;
    for i := 0 to (FCount-jumpsize0-2) do
      begin
        if SorCompare(FSorArray[i], FSorArray[SwapNumber]) > 0 then
          begin
            Temp := FSorArray[i];
            FSorArray[i] := FSorArray[SwapNumber];
            FSorArray[SwapNumber] := Temp;
            Result := false;
          end;
          inc(SwapNumber);
      end;
  until jumpsize0 = 0;
end;

{Driver for the " Comb " routine.
 Based on routines from the SWAG-Archive.
 Very fast, for a smaller number of items with large keys
 " Comb " may outperform Quicksort.
 ( Only a few thousends )}
procedure TSorArray.CombSortW(SorCompare: TSortArrayCompare);
begin
  comb(FCount, SorCompare);
  repeat
  until comb(0, SorCompare);
end;

{ Based on a non-recursive QuickSort from the SWAG-Archive.
  ( TV Sorting Unit by Brad Williams ) }
procedure TSorArray.QuickSort(SorCompare: TSortArrayCompare);
var Left, Right, SubArray, SubLeft, SubRight:LongInt;
    Temp, Pivot: Pointer;
    Stack : array[1..32] of record First, Last : LongInt; end;
begin
  SubArray := 1;
  Stack[SubArray].First := 0;
  Stack[SubArray].Last := Count - 1;
    repeat
      Left := Stack[SubArray].First;
      Right := Stack[SubArray].Last;
      Dec(SubArray);
      repeat
        SubLeft := Left;
        SubRight := Right;
        Pivot := FSorArray[(Left + Right) shr 1];
        repeat
          while SorCompare(FSorArray[SubLeft], Pivot) < 0 do Inc(SubLeft);
          while SorCompare(FSorArray[SubRight], Pivot) > 0 do Dec(SubRight);
          IF SubLeft <= SubRight then
            begin
              Temp := FSorArray[SubLeft];
              FSorArray[SubLeft] := FSorArray[SubRight];
              FSorArray[SubRight] := Temp;
              Inc(SubLeft);
              Dec(SubRight);
            end;
        until SubLeft > SubRight;
        IF SubLeft < Right then
          begin
            Inc(SubArray);
            Stack[SubArray].First := SubLeft;
            Stack[SubArray].Last := Right;
          end;
          Right := SubRight;
      until Left >= Right;
    until SubArray = 0;
end;

{Unfortunately the " Merge " routine needs additional memory
 An Algorithm to perform merging in linear time without extra space
 is described in:
 B. Huang and M. Langston, " Practical In-place Merging ",
 Communications of the ACM 31(1988), 348-352. }
procedure TSorArray.Merge(SorCompare: TSortArrayCompare);
var
  TempPos : integer;
begin
  TempPos := FLeftArray.FLeft;
  while ( FLeftArray.Full ) and ( FRightArray.Full ) do  { Main Loop }
    begin
      if SorCompare(FSorArray[FLeftArray.FLeft], FSorArray[FRightArray.FLeft]) <= 0 then
        begin
          TempArray[ TempPos ] := FSorArray[ FLeftArray.FLeft ];
          FLeftArray.Next;
        end
      else
        begin
          TempArray[ TempPos ] := FSorArray[ FRightArray.FLeft ];
          FRightArray.Next;
        end;
          inc(TempPos);
    end;

  while FLeftArray.Full do    { Copy Rest of First SubArray }
    begin
      TempArray[ TempPos ] := FSorArray[ FLeftArray.FLeft ];
      inc(TempPos); FLeftArray.Next;
    end;

  while FRightArray.Full do   { Copy Rest of Second SubArray }
    begin
      TempArray[ TempPos ] := FSorArray[ FRightArray.FLeft ];
      inc(TempPos); FRightArray.Next;
    end;
end;

{Non-recursive Mergesort.
 Very fast, if enough memory available.
 The number of comparisions used is nearly optimal, about 3/4 of QuickSort.
 If comparision plays a very more important role than exchangement,
 it outperforms QuickSort in any case.
 ( Large keys in pointer arrays, for example text with few short lines. )
 From all Algoritms with O(N lg N) it's the only stable, meaning it lefts
 equal keys in the order of input. This may be important in some cases. }
procedure TSorArray.MergeSort(SorCompare: TSortArrayCompare);
var
  a, b, N, todo: LongInt;
begin
  ReallocMem(TempArray, (FCount) * SizeOf(Pointer));
  FLeftArray:= TSubArray.Create(FCount -1);
  FRightArray:= TSubArray.Create(FCount -1);
  N:= 1;
  repeat
      todo:= 0;
      repeat
        a:= todo;
        b:= a +N;
        todo:= b +N;
        FLeftArray.Init(a, b -1);
        FRightArray.Init(b, todo -1);
        Merge(SorCompare);
      until todo >= Fcount;
      SwapArray:= FSorArray; {Alternating use of the arrays.}
      FSorArray:= TempArray;
      TempArray:= SwapArray;
      N:= N shl 1;
    until N >= Fcount;
    FLeftArray.Free;
    FRightArray.Free;
end;

end.
{--------------------------------------------------------------------}
{ Martin Waldenburg
  Landaeckerstrasse 27
  71642 Ludwigsburg
  Germany
  Share your Code}
{--------------------------------------------------------------------}
